import json

import pytest

# define a calendar event schema and then use a custom validator to validate that there
# are no events with "Occult" in their names
SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "definitions": {
        "calendar-event": {
            "type": "object",
            "properties": {
                "title": {"type": "string"},
                "start": {"type": "string", "format": "date-time"},
                "end": {"type": "string", "format": "date-time"},
            },
        }
    },
    "properties": {
        "events": {
            "type": "array",
            "items": {"$ref": "#/definitions/calendar-event"},
        },
    },
    "required": ["events"],
}
VALID_DOC = {
    "events": [
        {
            "title": "Weekly Production Meeting",
            "start": "2019-06-24T09:00:00-05:00",
            "end": "2019-06-24T10:00:00-05:00",
        },
        {
            "title": "Catch Up",
            "start": "2019-06-24T10:00:00-05:00",
            "end": "2019-06-24T10:30:00-05:00",
        },
    ]
}
INVALID_DOC = {
    "events": [
        {
            "title": "Weekly Production Meeting",
            "start": "2019-06-24T09:00:00-05:00",
            "end": "2019-06-24T10:00:00-05:00",
        },
        {
            "title": "Catch Up",
            "start": "2019-06-24T10:00:00-05:00",
            "end": "2019-06-24T10:30:00-05:00",
        },
        {
            "title": "Occult Study Session",
            "start": "2019-06-24T10:00:00-05:00",
            "end": "2019-06-24T12:00:00-05:00",
        },
    ]
}


@pytest.fixture(autouse=True)
def _foo_module(mock_module):
    mock_module(
        "foo.py",
        """\
import jsonschema

class MyValidator:
    def __init__(self, schema, *args, **kwargs):
        self.schema = schema
        self.real_validator = jsonschema.validators.Draft7Validator(
            schema, *args, **kwargs
        )

    def iter_errors(self, data, *args, **kwargs):
        yield from self.real_validator.iter_errors(data, *args, **kwargs)
        for event in data["events"]:
            if "Occult" in event["title"]:
                yield jsonschema.exceptions.ValidationError(
                    "Error! Occult event detected! Run!",
                    validator=None,
                    validator_value=None,
                    instance=event,
                    schema=self.schema,
                )
""",
    )


def test_custom_validator_class_can_detect_custom_conditions(run_line, tmp_path):
    doc = tmp_path / "invalid.json"
    doc.write_text(json.dumps(INVALID_DOC))

    schema = tmp_path / "schema.json"
    schema.write_text(json.dumps(SCHEMA))

    result = run_line(
        [
            "check-jsonschema",
            "--schemafile",
            str(schema),
            str(doc),
        ]
    )
    assert result.exit_code == 0, result.stdout  # pass

    result = run_line(
        [
            "check-jsonschema",
            "--schemafile",
            str(schema),
            "--validator-class",
            "foo:MyValidator",
            str(doc),
        ],
    )
    assert result.exit_code == 1  # fail
    assert "Occult event detected" in result.stdout, result.stdout


def test_custom_validator_class_can_pass_when_valid(run_line, tmp_path):
    doc = tmp_path / "valid.json"
    doc.write_text(json.dumps(VALID_DOC))

    schema = tmp_path / "schema.json"
    schema.write_text(json.dumps(SCHEMA))

    result = run_line(
        [
            "check-jsonschema",
            "--schemafile",
            str(schema),
            str(doc),
        ]
    )
    assert result.exit_code == 0, result.stdout  # pass

    result = run_line(
        [
            "check-jsonschema",
            "--schemafile",
            str(schema),
            "--validator-class",
            "foo:MyValidator",
            str(doc),
        ],
    )
    assert result.exit_code == 0  # pass



import os
import pathlib
import sys

import pytest
import responses


@pytest.fixture(autouse=True)
def mocked_responses():
    responses.start()
    yield
    responses.stop()
    responses.reset()


@pytest.fixture
def mock_module(tmp_path, monkeypatch):
    monkeypatch.syspath_prepend(tmp_path)
    all_names_to_clear = []

    def func(path, text):
        path = pathlib.Path(path)
        mod_dir = tmp_path / path.parent
        mod_dir.mkdir(parents=True, exist_ok=True)
        for part in path.parts[:-1]:
            (tmp_path / part / "__init__.py").touch()

        (tmp_path / path).write_text(text)

        for i in range(len(path.parts)):
            modname = ".".join(path.parts[: i + 1])
            if modname.endswith(".py"):
                modname = modname[:-3]
            all_names_to_clear.append(modname)

    yield func

    for name in all_names_to_clear:
        if name in sys.modules:
            del sys.modules[name]


@pytest.fixture
def in_tmp_dir(request, tmp_path):
    os.chdir(str(tmp_path))
    yield
    os.chdir(request.config.invocation_dir)



import random

import pytest

from check_jsonschema.formats.implementations.iso8601_time import validate


@pytest.mark.parametrize(
    "timestr",
    (
        "12:34:56Z",
        "23:59:59z",
        "23:59:59+00:00",
        "01:59:59-00:00",
    ),
)
def test_simple_positive_cases(timestr):
    assert validate(timestr)


@pytest.mark.parametrize(
    "timestr",
    (
        "12:34:56",
        "23:59:60Z",
        "23:59:59+24:00",
        "01:59:59-00:60",
        "01:01:00:00:60",
    ),
)
def test_simple_negative_cases(timestr):
    assert not validate(timestr)


@pytest.mark.parametrize("precision", list(range(20)))
@pytest.mark.parametrize(
    "offsetstr",
    (
        "Z",
        "+00:00",
        "-00:00",
        "+23:59",
    ),
)
def test_allows_fracsec(precision, offsetstr):
    fracsec = random.randint(0, 10**precision)
    assert validate(f"23:59:59.{fracsec}{offsetstr}")



import os
import platform

import pytest
from click.testing import CliRunner

from check_jsonschema.cli.main_command import build_checker
from check_jsonschema.cli.main_command import main as cli_main


@pytest.fixture
def runner() -> CliRunner:
    return CliRunner(mix_stderr=False)


@pytest.mark.skipif(
    platform.system() != "Linux", reason="test requires /proc/self/ mechanism"
)
def test_open_file_usage_never_exceeds_1000(runner, monkeypatch, tmp_path):
    schema_path = tmp_path / "schema.json"
    schema_path.write_text("{}")

    args = [
        "--schemafile",
        str(schema_path),
    ]

    for i in range(2000):
        instance_path = tmp_path / f"file{i}.json"
        instance_path.write_text("{}")
        args.append(str(instance_path))

    checker = None

    def fake_execute(argv):
        nonlocal checker
        checker = build_checker(argv)

    monkeypatch.setattr("check_jsonschema.cli.main_command.execute", fake_execute)
    res = runner.invoke(cli_main, args)
    assert res.exit_code == 0, res.stderr

    assert checker is not None
    assert len(os.listdir("/proc/self/fd")) < 2000
    for _fname, _data in checker._instance_loader.iter_files():
        assert len(os.listdir("/proc/self/fd")), 2000



from __future__ import annotations

import copy
import enum
import re
import typing as t

import jsonschema
import jsonschema.validators
import regress

from .implementations import validate_rfc3339, validate_time

# all known format strings except for a selection from draft3 which have either
# been renamed or removed:
# - color
# - host-name
# - ip-address
KNOWN_FORMATS: tuple[str, ...] = (
    "date",
    "date-time",
    "duration",
    "email",
    "hostname",
    "idn-email",
    "idn-hostname",
    "ipv4",
    "ipv6",
    "iri",
    "iri-reference",
    "json-pointer",
    "regex",
    "relative-json-pointer",
    "time",
    "uri",
    "uri-reference",
    "uri-template",
    "uuid",
)


class RegexVariantName(enum.Enum):
    default = "default"
    python = "python"


class RegexImplementation:
    def __init__(self, variant: RegexVariantName) -> None:
        self.variant = variant

    def check_format(self, instance: t.Any) -> bool:
        if not isinstance(instance, str):
            return True

        try:
            if self.variant == RegexVariantName.default:
                regress.Regex(instance)
            else:
                re.compile(instance)
        # something is wrong with RegressError getting into the published types
        # needs investigation... for now, ignore the error
        except (regress.RegressError, re.error):  # type: ignore[attr-defined]
            return False

        return True


class FormatOptions:
    def __init__(
        self,
        *,
        enabled: bool = True,
        regex_variant: RegexVariantName = RegexVariantName.default,
        disabled_formats: tuple[str, ...] = (),
    ) -> None:
        self.enabled = enabled
        self.regex_variant = regex_variant
        self.disabled_formats = disabled_formats


def get_base_format_checker(schema_dialect: str | None) -> jsonschema.FormatChecker:
    # resolve the dialect, if given, to a validator class
    # default to the latest draft
    validator_class = jsonschema.validators.validator_for(
        {} if schema_dialect is None else {"$schema": schema_dialect},
        default=jsonschema.Draft202012Validator,
    )
    return validator_class.FORMAT_CHECKER


def make_format_checker(
    opts: FormatOptions,
    schema_dialect: str | None = None,
) -> jsonschema.FormatChecker | None:
    if not opts.enabled:
        return None

    # copy the base checker
    base_checker = get_base_format_checker(schema_dialect)
    checker = copy.deepcopy(base_checker)

    # replace the regex check
    del checker.checkers["regex"]
    regex_impl = RegexImplementation(opts.regex_variant)
    checker.checks("regex")(regex_impl.check_format)
    checker.checks("date-time")(validate_rfc3339)
    checker.checks("time")(validate_time)

    # remove the disabled checks, which may include the regex check
    for checkname in opts.disabled_formats:
        if checkname not in checker.checkers:
            continue
        del checker.checkers[checkname]

    return checker



import re

# this regex is based on the one from the rfc3339-validator package
# credit to the original author
# original license:
#
#    MIT License
#
#    Copyright (c) 2019, Nicolas Aimetti
#
#    Permission is hereby granted, free of charge, to any person obtaining a copy
#    of this software and associated documentation files (the "Software"), to deal
#    in the Software without restriction, including without limitation the rights
#    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the Software is
#    furnished to do so, subject to the following conditions:
#
#    The above copyright notice and this permission notice shall be included in all
#    copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.
#
# modifications have been made for additional corner cases and speed
RFC3339_REGEX = re.compile(
    r"""
    ^
    (?:\d{4})
    -
    (?:0[1-9]|1[0-2])
    -
    (?:[0-3]\d)
    (?:[Tt])
    (?:[01]\d|2[0123])
    :
    (?:[0-5]\d)
    :
    (?:[0-5]\d)
    # (optional) fractional seconds
    (?:[\.,]\d+)?
    # UTC or offset
    (?:
        [Zz]
        | [+-](?:[01]\d|2[0123]):[0-5]\d
    )
    $
""",
    re.VERBOSE | re.ASCII,
)


def validate(date_str: object) -> bool:
    """Validate a string as a RFC3339 date-time."""
    if not isinstance(date_str, str):
        return False
    if not RFC3339_REGEX.match(date_str):
        return False

    year, month, day = int(date_str[:4]), int(date_str[5:7]), int(date_str[8:10])

    if month in {4, 6, 9, 11}:
        max_day = 30
    elif month == 2:
        max_day = 29 if year % 4 == 0 and (year % 100 != 0 or year % 400 == 0) else 28
    else:
        max_day = 31
    if not 1 <= day <= max_day:
        return False
    return True


if __name__ == "__main__":
    import timeit

    N = 100_000
    tests = (
        ("long_fracsec", "2018-12-31T23:59:59.8446519776713Z"),
        ("basic", "2018-12-31T23:59:59Z"),
        ("in_february", "2018-02-12T23:59:59Z"),
        ("in_february_invalid", "2018-02-29T23:59:59Z"),
        ("missing_t", "2018-12-31 23:59:59Z"),
        ("invalid_day", "2018-12-41T23:59:59Z"),
    )

    print("benchmarking")
    for name, val in tests:
        all_times = timeit.repeat(
            f"validate({val!r})", globals=globals(), repeat=3, number=N
        )
        print(f"{name} (valid={validate(val)}): {int(min(all_times) / N * 10**9)}ns")



import random

import pytest

from check_jsonschema.formats.implementations.rfc3339 import validate


@pytest.mark.parametrize(
    "datestr",
    (
        "2018-12-31T23:59:59Z",
        "2018-12-31t23:59:59Z",
        "2018-12-31t23:59:59z",
        "2018-12-31T23:59:59+00:00",
        "2018-12-31T23:59:59-00:00",
    ),
)
def test_simple_positive_cases(datestr):
    assert validate(datestr)


@pytest.mark.parametrize(
    "datestr",
    (
        "2018-12-31T23:59:59",
        "2018-12-31T23:59:59+00:00Z",
        "2018-12-31 23:59:59",
    ),
)
def test_simple_negative_case(datestr):
    assert not validate(datestr)


@pytest.mark.parametrize("precision", list(range(20)))
@pytest.mark.parametrize(
    "offsetstr",
    (
        "Z",
        "+00:00",
        "-00:00",
        "+23:59",
    ),
)
def test_allows_fracsec(precision, offsetstr):
    fracsec = random.randint(0, 10**precision)
    assert validate(f"2018-12-31T23:59:59.{fracsec}{offsetstr}")


@pytest.mark.parametrize(
    "datestr",
    (
        # no such month
        "2020-13-01T00:00:00Z",
        "2020-00-01T00:00:00Z",
        # no such day
        "2020-01-00T00:00:00Z",
        "2020-01-32T00:00:00Z",
    ),
)
def test_basic_bounds_validated(datestr):
    assert not validate(datestr)


@pytest.mark.parametrize(
    "month, maxday",
    (
        (1, 31),
        (3, 31),
        (4, 30),
        (5, 31),
        (6, 30),
        (7, 31),
        (8, 31),
        (9, 30),
        (10, 31),
        (11, 30),
    ),
)
def test_day_bounds_by_month(month, maxday):
    good_date = f"2020-{month:02}-{maxday:02}T00:00:00Z"
    bad_date = f"2020-{month:02}-{maxday+1:02}T00:00:00Z"
    assert validate(good_date)
    assert not validate(bad_date)


@pytest.mark.parametrize(
    "year, maxday",
    (
        (2018, 28),
        (2016, 29),
        (2400, 29),
        (2500, 28),
    ),
)
def test_day_bounds_for_february(year, maxday):
    good_date = f"{year}-02-{maxday:02}T00:00:00Z"
    bad_date = f"{year}-02-{maxday+1:02}T00:00:00Z"
    assert validate(good_date)
    assert not validate(bad_date)



import json

import pytest

from check_jsonschema.parsers.json5 import ENABLED as JSON5_ENABLED

SIMPLE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "properties": {
        "title": {"type": "string"},
    },
    "additionalProperties": False,
}
YAML_REF_MAIN_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "properties": {
        "title": {"$ref": "./title_schema.yaml"},
    },
    "additionalProperties": False,
}
JSON5_REF_MAIN_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "properties": {
        "title": {"$ref": "./title_schema.json5"},
    },
    "additionalProperties": False,
}
TITLE_SCHEMA = {"type": "string"}
PASSING_DOCUMENT = {"title": "doc one"}
FAILING_DOCUMENT = {"title": 2}


@pytest.mark.parametrize("passing_data", [True, False])
def test_yaml_reference(run_line, tmp_path, passing_data):
    main_schemafile = tmp_path / "main_schema.json"
    main_schemafile.write_text(json.dumps(YAML_REF_MAIN_SCHEMA))
    # JSON is a subset of YAML, so this works for generated YAML
    ref_schema = tmp_path / "title_schema.yaml"
    ref_schema.write_text(json.dumps(TITLE_SCHEMA))

    doc = tmp_path / "doc.json"
    if passing_data:
        doc.write_text(json.dumps(PASSING_DOCUMENT))
    else:
        doc.write_text(json.dumps(FAILING_DOCUMENT))

    result = run_line(
        ["check-jsonschema", "--schemafile", str(main_schemafile), str(doc)]
    )
    assert result.exit_code == (0 if passing_data else 1)


@pytest.mark.skipif(not JSON5_ENABLED, reason="test requires json5")
@pytest.mark.parametrize("passing_data", [True, False])
def test_json5_reference(run_line, tmp_path, passing_data):
    main_schemafile = tmp_path / "main_schema.json"
    main_schemafile.write_text(json.dumps(JSON5_REF_MAIN_SCHEMA))
    ref_schema = tmp_path / "title_schema.json5"
    ref_schema.write_text(json.dumps(TITLE_SCHEMA))

    doc = tmp_path / "doc.json"
    if passing_data:
        doc.write_text(json.dumps(PASSING_DOCUMENT))
    else:
        doc.write_text(json.dumps(FAILING_DOCUMENT))

    result = run_line(
        ["check-jsonschema", "--schemafile", str(main_schemafile), str(doc)]
    )
    assert result.exit_code == (0 if passing_data else 1)


@pytest.mark.skipif(not JSON5_ENABLED, reason="test requires json5")
@pytest.mark.parametrize("passing_data", [True, False])
def test_can_load_json5_schema(run_line, tmp_path, passing_data):
    # dump JSON to the JSON5 file, this is fine
    main_schemafile = tmp_path / "main_schema.json5"
    main_schemafile.write_text(json.dumps(SIMPLE_SCHEMA))

    doc = tmp_path / "doc.json"
    if passing_data:
        doc.write_text(json.dumps(PASSING_DOCUMENT))
    else:
        doc.write_text(json.dumps(FAILING_DOCUMENT))

    result = run_line(
        ["check-jsonschema", "--schemafile", str(main_schemafile), str(doc)]
    )
    assert result.exit_code == (0 if passing_data else 1)



from __future__ import annotations

import pathlib

import jsonschema

from .parsers import ParseError


class CheckResult:
    def __init__(self) -> None:
        self.validation_errors: dict[str, list[jsonschema.ValidationError]] = {}
        self.parse_errors: dict[str, list[ParseError]] = {}
        self.successes: list[str] = []

    @property
    def success(self) -> bool:
        return not (bool(self.parse_errors) or bool(self.validation_errors))

    def record_validation_success(self, path: pathlib.Path | str) -> None:
        self.successes.append(str(path))

    def record_validation_error(
        self, path: pathlib.Path | str, err: jsonschema.ValidationError
    ) -> None:
        filename = str(path)
        if filename not in self.validation_errors:
            self.validation_errors[filename] = []
        self.validation_errors[filename].append(err)

    def record_parse_error(self, path: pathlib.Path | str, err: ParseError) -> None:
        filename = str(path)
        if filename not in self.parse_errors:
            self.parse_errors[filename] = []
        self.parse_errors[filename].append(err)



from __future__ import annotations

import io
import pathlib
import typing as t

import ruamel.yaml

from ..identify_filetype import path_to_type
from . import json5, json_, toml, yaml

_PARSER_ERRORS: set[type[Exception]] = {
    json_.JSONDecodeError,
    yaml.ParseError,
    toml.ParseError,
}
DEFAULT_LOAD_FUNC_BY_TAG: dict[str, t.Callable[[t.IO[bytes]], t.Any]] = {
    "json": json_.load,
    "toml": toml.load,
}
SUPPORTED_FILE_FORMATS = ["json", "toml", "yaml"]
if json5.ENABLED:
    SUPPORTED_FILE_FORMATS.append("json5")
    DEFAULT_LOAD_FUNC_BY_TAG["json5"] = json5.load
    _PARSER_ERRORS.add(json5.ParseError)
MISSING_SUPPORT_MESSAGES: dict[str, str] = {
    "json5": json5.MISSING_SUPPORT_MESSAGE,
}
LOADING_FAILURE_ERROR_TYPES: tuple[type[Exception], ...] = tuple(_PARSER_ERRORS)


class ParseError(ValueError):
    pass


class BadFileTypeError(ParseError):
    pass


class FailedFileLoadError(ParseError):
    pass


class ParserSet:
    def __init__(
        self,
        *,
        modify_yaml_implementation: t.Callable[[ruamel.yaml.YAML], None] | None = None,
        supported_formats: t.Sequence[str] | None = None,
    ) -> None:
        yaml_impl = yaml.construct_yaml_implementation()
        failover_yaml_impl = yaml.construct_yaml_implementation(pure=True)
        if modify_yaml_implementation:
            modify_yaml_implementation(yaml_impl)
            modify_yaml_implementation(failover_yaml_impl)
        base_by_tag = {
            "yaml": yaml.impl2loader(yaml_impl, failover_yaml_impl),
            **DEFAULT_LOAD_FUNC_BY_TAG,
        }
        if supported_formats is None:
            self._by_tag = base_by_tag
        else:
            self._by_tag = {
                k: v for k, v in base_by_tag.items() if k in supported_formats
            }

    def get(
        self, path: pathlib.Path | str, default_filetype: str
    ) -> t.Callable[[t.IO[bytes]], t.Any]:
        filetype = path_to_type(path, default_type=default_filetype)

        if filetype in self._by_tag:
            return self._by_tag[filetype]

        if filetype in MISSING_SUPPORT_MESSAGES:
            raise BadFileTypeError(
                f"cannot parse {path} because support is missing for {filetype}\n"
                + MISSING_SUPPORT_MESSAGES[filetype]
            )
        raise BadFileTypeError(
            f"cannot parse {path} as it is not one of the supported filetypes: "
            + ",".join(self._by_tag.keys())
        )

    def parse_data_with_path(
        self, data: t.IO[bytes] | bytes, path: pathlib.Path | str, default_filetype: str
    ) -> t.Any:
        loadfunc = self.get(path, default_filetype)
        try:
            if isinstance(data, bytes):
                data = io.BytesIO(data)
            return loadfunc(data)
        except LOADING_FAILURE_ERROR_TYPES as e:
            raise FailedFileLoadError(f"Failed to parse {path}") from e

    def parse_file(self, path: pathlib.Path | str, default_filetype: str) -> t.Any:
        with open(path, "rb") as fp:
            return self.parse_data_with_path(fp, path, default_filetype)



from __future__ import annotations

import contextlib
import io
import os
import platform
import shutil
import tempfile
import time
import typing as t

import requests


class FailedDownloadError(Exception):
    pass


class CacheDownloader:
    _LASTMOD_FMT = "%a, %d %b %Y %H:%M:%S %Z"

    # changed in v0.5.0
    # original cache dir was "jsonschema_validate"
    # this will let us do any other caching we might need in the future in the same
    # cache dir (adjacent to "downloads")
    _CACHEDIR_NAME = os.path.join("check_jsonschema", "downloads")

    def __init__(
        self,
        file_url: str,
        filename: str | None = None,
        cache_dir: str | None = None,
        disable_cache: bool = False,
        validation_callback: t.Callable[[bytes], t.Any] | None = None,
    ):
        self._file_url = file_url
        self._filename = filename or file_url.split("/")[-1]
        self._cache_dir = cache_dir or self._compute_default_cache_dir()
        self._disable_cache = disable_cache
        self._validation_callback = validation_callback

    def _compute_default_cache_dir(self) -> str | None:
        sysname = platform.system()

        # on windows, try to get the appdata env var
        # this *could* result in cache_dir=None, which is fine, just skip caching in
        # that case
        if sysname == "Windows":
            cache_dir = os.getenv("LOCALAPPDATA", os.getenv("APPDATA"))
        # macOS -> app support dir
        elif sysname == "Darwin":
            cache_dir = os.path.expanduser("~/Library/Caches")
        # default for unknown platforms, namely linux behavior
        # use XDG env var and default to ~/.cache/
        else:
            cache_dir = os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache"))

        if cache_dir:
            cache_dir = os.path.join(cache_dir, self._CACHEDIR_NAME)

        return cache_dir

    def _get_request(self) -> requests.Response:
        try:
            # do manual retries, rather than using urllib3 retries, to make it trivially
            # testable with 'responses'
            r: requests.Response | None = None
            for _attempt in range(3):
                r = requests.get(self._file_url, stream=True)
                if r.ok:
                    if self._validation_callback is not None:
                        try:
                            self._validation_callback(r.content)
                        except ValueError:
                            continue
                    return r
            assert r is not None
            raise FailedDownloadError(
                f"got responses with status={r.status_code}, retries exhausted"
            )
        except requests.RequestException as e:
            raise FailedDownloadError("encountered error during download") from e

    def _lastmod_from_response(self, response: requests.Response) -> float:
        try:
            return time.mktime(
                time.strptime(response.headers["last-modified"], self._LASTMOD_FMT)
            )
        # OverflowError: time outside of platform-specific bounds
        # ValueError: malformed/unparseable
        # LookupError: no such header
        except (OverflowError, ValueError, LookupError):
            return 0.0

    def _cache_hit(self, cachefile: str, response: requests.Response) -> bool:
        # no file? miss
        if not os.path.exists(cachefile):
            return False

        # compare mtime on any cached file against the remote last-modified time
        # it is considered a hit if the local file is at least as new as the remote file
        local_mtime = os.path.getmtime(cachefile)
        remote_mtime = self._lastmod_from_response(response)
        return local_mtime >= remote_mtime

    def _write(self, dest: str, response: requests.Response) -> None:
        # download to a temp file and then move to the dest
        # this makes the download safe if run in parallel (parallel runs
        # won't create a new empty file for writing and cause failures)
        fp = tempfile.NamedTemporaryFile(mode="wb", delete=False)
        fp.write(response.content)
        fp.close()
        shutil.copy(fp.name, dest)
        os.remove(fp.name)

    def _download(self) -> str:
        assert self._cache_dir
        os.makedirs(self._cache_dir, exist_ok=True)
        dest = os.path.join(self._cache_dir, self._filename)

        response = self._get_request()
        # check to see if we have a file which matches the connection
        # only download if we do not (cache miss, vs hit)
        if not self._cache_hit(dest, response):
            self._write(dest, response)

        return dest

    @contextlib.contextmanager
    def open(self) -> t.Iterator[t.IO[bytes]]:
        if (not self._cache_dir) or self._disable_cache:
            yield io.BytesIO(self._get_request().content)
        else:
            with open(self._download(), "rb") as fp:
                yield fp



from __future__ import annotations

import dataclasses
import importlib.util
import shlex
from pathlib import Path

import pytest
import ruamel.yaml

from check_jsonschema.parsers.json5 import ENABLED as JSON5_ENABLED

yaml = ruamel.yaml.YAML(typ="safe")

HERE = Path(__file__).parent
EXAMPLE_FILES = HERE.parent / "example-files"
EXAMPLE_HOOK_FILES = HERE.parent / "example-files" / "hooks"
EXAMPLE_EXPLICIT_FILES = HERE.parent / "example-files" / "explicit-schema"


def _iter_hook_config():
    config_file = HERE.parent.parent / ".pre-commit-hooks.yaml"
    with open(config_file) as fp:
        for hook in yaml.load(fp):
            hookid = hook["id"]
            if not hookid.startswith("check-"):
                continue
            hookid = hookid[len("check-") :]
            entry = shlex.split(hook["entry"])
            yield (hookid, entry)


HOOK_CONFIG = dict(_iter_hook_config())


def _build_hook_cases(category):
    res = {}
    for hookid in HOOK_CONFIG:
        example_dir = EXAMPLE_HOOK_FILES / category / hookid
        if example_dir.exists():
            for example in example_dir.iterdir():
                if example.name == "_config.yaml":
                    continue
                res[str(example.relative_to(EXAMPLE_HOOK_FILES / category))] = hookid
    return res


def _get_explicit_cases(category):
    res = []
    example_dir = EXAMPLE_EXPLICIT_FILES / category
    for example in example_dir.iterdir():
        res.append(str(example.relative_to(EXAMPLE_EXPLICIT_FILES / category)))
    return res


POSITIVE_HOOK_CASES = _build_hook_cases("positive")
NEGATIVE_HOOK_CASES = _build_hook_cases("negative")


@pytest.mark.parametrize("case_name", POSITIVE_HOOK_CASES.keys())
def test_hook_positive_examples(case_name, run_line):
    rcase = ResolvedCase.load_positive(case_name)

    hook_id = POSITIVE_HOOK_CASES[case_name]
    ret = run_line(HOOK_CONFIG[hook_id] + [rcase.path] + rcase.add_args)
    assert ret.exit_code == 0, _format_cli_result(rcase, ret)


@pytest.mark.parametrize("case_name", NEGATIVE_HOOK_CASES.keys())
def test_hook_negative_examples(case_name, run_line):
    rcase = ResolvedCase.load_negative(case_name)

    hook_id = NEGATIVE_HOOK_CASES[case_name]
    ret = run_line(HOOK_CONFIG[hook_id] + [rcase.path] + rcase.add_args)
    assert ret.exit_code == 1, _format_cli_result(rcase, ret)


@pytest.mark.parametrize("case_name", _get_explicit_cases("positive"))
def test_explicit_positive_examples(case_name, run_line):
    _check_file_format_skip(case_name)
    casedir = EXAMPLE_EXPLICIT_FILES / "positive" / case_name

    instance = casedir / "instance.json"
    if not instance.exists():
        instance = casedir / "instance.yaml"
    if not instance.exists():
        instance = casedir / "instance.toml"
    if not instance.exists():
        raise Exception("could not find an instance file for test case")

    schema = casedir / "schema.json"
    if not schema.exists():
        schema = casedir / "schema.yaml"
    if not schema.exists():
        raise Exception("could not find a schema file for test case")

    ret = run_line(
        [
            "check-jsonschema",
            "--schemafile",
            str(schema),
            str(instance),
        ]
    )
    assert ret.exit_code == 0


def _check_file_format_skip(case_name):
    if case_name.endswith("json5") and not JSON5_ENABLED:
        pytest.skip("cannot check json5 support without json5 enabled")


@dataclasses.dataclass
class ResolvedCase:
    category: str
    path: str
    add_args: list[str]
    config: dict

    def check_skip(self) -> None:
        if "requires_packages" in self.config:
            for pkg in self.config["requires_packages"]:
                if _package_is_installed(pkg):
                    continue
                pytest.skip(f"cannot check because '{pkg}' is not installed")

    def __post_init__(self) -> None:
        self.check_skip()

    @classmethod
    def load_positive(cls: type[ResolvedCase], case_name: str) -> ResolvedCase:
        return cls._load("positive", case_name)

    @classmethod
    def load_negative(cls: type[ResolvedCase], case_name: str) -> ResolvedCase:
        return cls._load("negative", case_name)

    @classmethod
    def _load(cls: type[ResolvedCase], category: str, case_name: str) -> ResolvedCase:
        _check_file_format_skip(case_name)

        path = EXAMPLE_HOOK_FILES / category / case_name
        config = cls._load_file_config(path.parent / "_config.yaml", path.name)

        return cls(
            category=category,
            path=str(path),
            add_args=config.get("add_args", []),
            config=config,
        )

    @staticmethod
    def _load_file_config(config_path, name):
        if not config_path.is_file():
            return {}
        with open(config_path) as fp:
            loaded_conf = yaml.load(fp)
        files_section = loaded_conf.get("files", {})
        return files_section.get(name, {})


def _package_is_installed(pkg: str) -> bool:
    spec = importlib.util.find_spec(pkg)
    if spec is None:
        return False
    return True


def _format_cli_result(rcase: ResolvedCase, result) -> str:
    return (
        "\n"
        f"config.add_args={rcase.add_args}\n"
        f"{result.exit_code=}\n"
        f"result.stdout={result.output}\n"
        f"{result.stderr=}"
    )



import json

import pytest
import responses

from check_jsonschema import cachedownloader

CASES = {
    "case1": {
        "main_schema": {
            "$schema": "http://json-schema.org/draft-07/schema",
            "properties": {
                "title": {"$ref": "./title_schema.json"},
            },
            "additionalProperties": False,
        },
        "other_schemas": {"title_schema": {"type": "string"}},
        "passing_document": {"title": "doc one"},
        "failing_document": {"title": 2},
    },
    "case2": {
        "main_schema": {
            "$schema": "http://json-schema.org/draft-07/schema",
            "type": "object",
            "required": ["test"],
            "properties": {"test": {"$ref": "./values.json#/$defs/test"}},
        },
        "other_schemas": {
            "values": {
                "$schema": "http://json-schema.org/draft-07/schema",
                "$defs": {"test": {"type": "string"}},
            }
        },
        "passing_document": {"test": "some data"},
        "failing_document": {"test": {"foo": "bar"}},
    },
}


@pytest.fixture(autouse=True)
def _mock_schema_cache_dir(monkeypatch, tmp_path):
    def _fake_compute_default_cache_dir(self):
        return str(tmp_path)

    monkeypatch.setattr(
        cachedownloader.CacheDownloader,
        "_compute_default_cache_dir",
        _fake_compute_default_cache_dir,
    )


@pytest.mark.parametrize("check_passes", (True, False))
@pytest.mark.parametrize("casename", ("case1", "case2"))
def test_remote_ref_resolution_simple_case(run_line, check_passes, casename, tmp_path):
    main_schema_loc = "https://example.com/main.json"
    responses.add("GET", main_schema_loc, json=CASES[casename]["main_schema"])
    for name, subschema in CASES[casename]["other_schemas"].items():
        other_schema_loc = f"https://example.com/{name}.json"
        responses.add("GET", other_schema_loc, json=subschema)

    instance_path = tmp_path / "instance.json"
    instance_path.write_text(
        json.dumps(
            CASES[casename]["passing_document"]
            if check_passes
            else CASES[casename]["failing_document"]
        )
    )

    result = run_line(
        ["check-jsonschema", "--schemafile", main_schema_loc, str(instance_path)]
    )
    output = f"\nstdout:\n{result.stdout}\n\nstderr:\n{result.stderr}"
    if check_passes:
        assert result.exit_code == 0, output
    else:
        assert result.exit_code == 1, output


# this test ensures that `$id` is preferred for the base URI over
# the retrieval URI
@pytest.mark.parametrize("check_passes", (True, False))
def test_ref_resolution_prefers_id_over_retrieval_uri(run_line, tmp_path, check_passes):
    main_schema = {
        "$id": "https://example.org/schemas/main.json",
        "$schema": "http://json-schema.org/draft-07/schema",
        "properties": {
            "title": {"$ref": "./title_schema.json"},
        },
        "additionalProperties": False,
    }
    title_schema = {"type": "string"}

    retrieval_uri = "https://example.org/alternate-path-retrieval-only/schemas/main"
    responses.add("GET", retrieval_uri, json=main_schema)
    responses.add(
        "GET", "https://example.org/schemas/title_schema.json", json=title_schema
    )

    instance_path = tmp_path / "instance.json"
    instance_path.write_text(json.dumps({"title": "doc one" if check_passes else 2}))

    result = run_line(
        ["check-jsonschema", "--schemafile", retrieval_uri, str(instance_path)]
    )
    output = f"\nstdout:\n{result.stdout}\n\nstderr:\n{result.stderr}"
    if check_passes:
        assert result.exit_code == 0, output
    else:
        assert result.exit_code == 1, output


@pytest.mark.parametrize("check_passes", (True, False))
def test_ref_resolution_does_not_callout_for_absolute_ref_to_retrieval_uri(
    run_line, tmp_path, check_passes
):
    retrieval_uri = "https://example.net/schemas/main"

    main_schema = {
        "$id": "https://example.net/schemas/some-uri-which-will-never-be-used/main.json",
        "$schema": "http://json-schema.org/draft-07/schema",
        "$defs": {"title": {"type": "string"}},
        "properties": {
            "title": {"$ref": f"{retrieval_uri}#/$defs/title"},
        },
        "additionalProperties": False,
    }

    # exactly one GET to the retrieval URI will work
    responses.add("GET", retrieval_uri, json=main_schema)
    responses.add("GET", retrieval_uri, json={"error": "permafrost melted"}, status=500)

    instance_path = tmp_path / "instance.json"
    instance_path.write_text(json.dumps({"title": "doc one" if check_passes else 2}))

    result = run_line(
        ["check-jsonschema", "--schemafile", retrieval_uri, str(instance_path)]
    )
    output = f"\nstdout:\n{result.stdout}\n\nstderr:\n{result.stderr}"
    if check_passes:
        assert result.exit_code == 0, output
    else:
        assert result.exit_code == 1, output


# this test ensures that `$id` is overwritten when `--base-uri` is used
@pytest.mark.parametrize("check_passes", (True, False))
def test_ref_resolution_with_custom_base_uri(run_line, tmp_path, check_passes):
    retrieval_uri = "https://example.org/retrieval-and-in-schema-only/schemas/main"
    explicit_base_uri = "https://example.org/schemas/main"
    main_schema = {
        "$id": retrieval_uri,
        "$schema": "http://json-schema.org/draft-07/schema",
        "properties": {
            "title": {"$ref": "./title_schema.json"},
        },
        "additionalProperties": False,
    }
    title_schema = {"type": "string"}

    responses.add("GET", retrieval_uri, json=main_schema)
    responses.add(
        "GET", "https://example.org/schemas/title_schema.json", json=title_schema
    )

    instance_path = tmp_path / "instance.json"
    instance_path.write_text(json.dumps({"title": "doc one" if check_passes else 2}))

    result = run_line(
        [
            "check-jsonschema",
            "--schemafile",
            retrieval_uri,
            "--base-uri",
            explicit_base_uri,
            str(instance_path),
        ]
    )
    output = f"\nstdout:\n{result.stdout}\n\nstderr:\n{result.stderr}"
    if check_passes:
        assert result.exit_code == 0, output
    else:
        assert result.exit_code == 1, output



def test_checker_non_json_schemafile(run_line, tmp_path):
    foo = tmp_path / "foo.json"
    bar = tmp_path / "bar.json"
    foo.write_text("{")
    bar.write_text("{}")

    res = run_line(["check-jsonschema", "--schemafile", str(foo), str(bar)])
    assert res.exit_code == 1
    assert "schemafile could not be parsed" in res.stderr


def test_checker_invalid_schemafile(run_line, tmp_path):
    foo = tmp_path / "foo.json"
    bar = tmp_path / "bar.json"
    foo.write_text('{"title": {"foo": "bar"}}')
    bar.write_text("{}")

    res = run_line(["check-jsonschema", "--schemafile", str(foo), str(bar)])
    assert res.exit_code == 1
    assert "schemafile was not valid" in res.stderr


def test_checker_invalid_schemafile_scheme(run_line, tmp_path):
    foo = tmp_path / "foo.json"
    bar = tmp_path / "bar.json"
    foo.write_text('{"title": "foo"}')
    bar.write_text("{}")

    res = run_line(["check-jsonschema", "--schemafile", f"ftp://{foo}", str(bar)])
    assert res.exit_code == 1
    assert "only supports http, https" in res.stderr



from __future__ import annotations

import typing as t
import warnings

import click


def deprecation_warning_callback(
    optstring: str, *, is_flag: bool = False, append_message: str | None = None
) -> t.Callable[[click.Context, click.Parameter, t.Any], t.Any]:
    def callback(ctx: click.Context, param: click.Parameter, value: t.Any) -> t.Any:
        if not value:
            return value
        if (is_flag and bool(value) is True) or (value is not None):
            message = (
                f"'{optstring}' is deprecated and will be removed in a future release."
            )
            if append_message is not None:
                message += f" {append_message}"
            warnings.warn(message, stacklevel=2)

        return value

    return callback



from __future__ import annotations

import pathlib
import typing as t
import urllib.error
import urllib.parse

import jsonschema

from ..builtin_schemas import get_builtin_schema
from ..formats import FormatOptions, make_format_checker
from ..parsers import ParserSet
from ..utils import is_url_ish
from .errors import UnsupportedUrlScheme
from .readers import HttpSchemaReader, LocalSchemaReader, StdinSchemaReader
from .resolver import make_reference_registry


def _extend_with_default(
    validator_class: type[jsonschema.protocols.Validator],
) -> type[jsonschema.Validator]:
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults_then_validate(
        validator: jsonschema.Validator,
        properties: dict[str, dict[str, t.Any]],
        instance: dict[str, t.Any],
        schema: dict[str, t.Any],
    ) -> t.Iterator[jsonschema.ValidationError]:
        for property_name, subschema in properties.items():
            if "default" in subschema and property_name not in instance:
                instance[property_name] = subschema["default"]

        yield from validate_properties(
            validator,
            properties,
            instance,
            schema,
        )

    return jsonschema.validators.extend(
        validator_class,
        {"properties": set_defaults_then_validate},
    )


class SchemaLoaderBase:
    def get_validator(
        self,
        path: pathlib.Path | str,
        instance_doc: dict[str, t.Any],
        format_opts: FormatOptions,
        fill_defaults: bool,
    ) -> jsonschema.protocols.Validator:
        raise NotImplementedError


class SchemaLoader(SchemaLoaderBase):
    validator_class: type[jsonschema.protocols.Validator] | None = None

    def __init__(
        self,
        schemafile: str,
        cache_filename: str | None = None,
        disable_cache: bool = False,
        base_uri: str | None = None,
        validator_class: type[jsonschema.protocols.Validator] | None = None,
    ) -> None:
        # record input parameters (these are not to be modified)
        self.schemafile = schemafile
        self.cache_filename = cache_filename
        self.disable_cache = disable_cache
        self.base_uri = base_uri
        self.validator_class = validator_class

        # if the schema location is a URL, which may include a file:// URL, parse it
        self.url_info = None
        if is_url_ish(self.schemafile):
            self.url_info = urllib.parse.urlparse(self.schemafile)

        # setup a parser collection
        self._parsers = ParserSet()

        # setup a schema reader lazily, when needed
        self._reader: (
            LocalSchemaReader | HttpSchemaReader | StdinSchemaReader | None
        ) = None

    @property
    def reader(self) -> LocalSchemaReader | HttpSchemaReader | StdinSchemaReader:
        if self._reader is None:
            self._reader = self._get_schema_reader()
        return self._reader

    def _get_schema_reader(
        self,
    ) -> LocalSchemaReader | HttpSchemaReader | StdinSchemaReader:
        if self.schemafile == "-":
            return StdinSchemaReader()

        if self.url_info is None or self.url_info.scheme in ("file", ""):
            return LocalSchemaReader(self.schemafile)

        if self.url_info.scheme in ("http", "https"):
            return HttpSchemaReader(
                self.schemafile,
                self.cache_filename,
                self.disable_cache,
            )
        else:
            raise UnsupportedUrlScheme(
                "check-jsonschema only supports http, https, and local files. "
                f"detected parsed URL had an unrecognized scheme: {self.url_info}"
            )

    def get_schema_retrieval_uri(self) -> str | None:
        return self.reader.get_retrieval_uri()

    def get_schema(self) -> dict[str, t.Any]:
        data = self.reader.read_schema()
        if self.base_uri is not None:
            data["$id"] = self.base_uri
        return data

    def get_validator(
        self,
        path: pathlib.Path | str,
        instance_doc: dict[str, t.Any],
        format_opts: FormatOptions,
        fill_defaults: bool,
    ) -> jsonschema.protocols.Validator:
        retrieval_uri = self.get_schema_retrieval_uri()
        schema = self.get_schema()

        schema_dialect = schema.get("$schema")

        # format checker (which may be None)
        format_checker = make_format_checker(format_opts, schema_dialect)

        # reference resolution
        # with support for YAML, TOML, and other formats from the parsers
        reference_registry = make_reference_registry(
            self._parsers, retrieval_uri, schema
        )

        if self.validator_class is None:
            # get the correct validator class and check the schema under its metaschema
            validator_cls = jsonschema.validators.validator_for(schema)
            validator_cls.check_schema(schema)
        else:
            # for a user-provided validator class, don't check_schema
            # on the grounds that it might *not* be valid but the user wants to use
            # their custom validator anyway
            #
            # in fact, there's no real guarantee that a user-provided
            # validator_class properly conforms to the jsonschema.Validator protocol
            # we *hope* that it does, but we can't be fully sure
            validator_cls = self.validator_class

        # extend the validator class with default-filling behavior if appropriate
        if fill_defaults:
            validator_cls = _extend_with_default(validator_cls)

        # now that we know it's safe to try to create the validator instance, do it
        validator = validator_cls(
            schema,
            registry=reference_registry,
            format_checker=format_checker,
        )
        return t.cast(jsonschema.protocols.Validator, validator)


class BuiltinSchemaLoader(SchemaLoader):
    def __init__(self, schema_name: str, base_uri: str | None = None) -> None:
        self.schema_name = schema_name
        self.base_uri = base_uri
        self._parsers = ParserSet()

    def get_schema_retrieval_uri(self) -> str | None:
        return None

    def get_schema(self) -> dict[str, t.Any]:
        data = get_builtin_schema(self.schema_name)
        if self.base_uri is not None:
            data["$id"] = self.base_uri
        return data


class MetaSchemaLoader(SchemaLoaderBase):
    def __init__(self, base_uri: str | None = None) -> None:
        if base_uri is not None:
            raise NotImplementedError(
                "'--base-uri' was used with '--metaschema'. "
                "This combination is not supported."
            )

    def get_validator(
        self,
        path: pathlib.Path | str,
        instance_doc: dict[str, t.Any],
        format_opts: FormatOptions,
        fill_defaults: bool,
    ) -> jsonschema.protocols.Validator:
        schema_validator = jsonschema.validators.validator_for(instance_doc)
        meta_validator_class = jsonschema.validators.validator_for(
            schema_validator.META_SCHEMA, default=schema_validator
        )

        # format checker (which may be None)
        meta_schema_dialect = schema_validator.META_SCHEMA.get("$schema")
        format_checker = make_format_checker(format_opts, meta_schema_dialect)

        meta_validator = meta_validator_class(
            schema_validator.META_SCHEMA, format_checker=format_checker
        )
        return meta_validator



from __future__ import annotations

import pathlib
import typing as t

import click
import jsonschema
import referencing.exceptions

from . import utils
from .formats import FormatOptions
from .instance_loader import InstanceLoader
from .parsers import ParseError
from .reporter import Reporter
from .result import CheckResult
from .schema_loader import SchemaLoaderBase, SchemaParseError, UnsupportedUrlScheme


class _Exit(Exception):
    def __init__(self, code: int):
        self.code = code


class SchemaChecker:
    def __init__(
        self,
        schema_loader: SchemaLoaderBase,
        instance_loader: InstanceLoader,
        reporter: Reporter,
        *,
        format_opts: FormatOptions | None = None,
        traceback_mode: str = "short",
        fill_defaults: bool = False,
    ):
        self._schema_loader = schema_loader
        self._instance_loader = instance_loader
        self._reporter = reporter

        self._format_opts = format_opts if format_opts is not None else FormatOptions()
        self._traceback_mode = traceback_mode
        self._fill_defaults = fill_defaults

    def _fail(self, msg: str, err: Exception | None = None) -> t.NoReturn:
        click.echo(msg, err=True)
        if err is not None:
            utils.print_error(err, mode=self._traceback_mode)
        raise _Exit(1)

    def get_validator(
        self, path: pathlib.Path | str, doc: dict[str, t.Any]
    ) -> jsonschema.protocols.Validator:
        try:
            return self._schema_loader.get_validator(
                path, doc, self._format_opts, self._fill_defaults
            )
        except SchemaParseError as e:
            self._fail("Error: schemafile could not be parsed as JSON", e)
        except jsonschema.SchemaError as e:
            self._fail(f"Error: schemafile was not valid: {e}\n", e)
        except UnsupportedUrlScheme as e:
            self._fail(f"Error: {e}\n", e)
        except Exception as e:
            self._fail("Error: Unexpected Error building schema validator", e)

    def _build_result(self) -> CheckResult:
        result = CheckResult()
        for path, data in self._instance_loader.iter_files():
            if isinstance(data, ParseError):
                result.record_parse_error(path, data)
            else:
                validator = self.get_validator(path, data)
                passing = True
                for err in validator.iter_errors(data):
                    result.record_validation_error(path, err)
                    passing = False
                if passing:
                    result.record_validation_success(path)
        return result

    def _run(self) -> None:
        try:
            result = self._build_result()
        except (
            referencing.exceptions.NoSuchResource,
            referencing.exceptions.Unretrievable,
            referencing.exceptions.Unresolvable,
        ) as e:
            self._fail("Failure resolving $ref within schema\n", e)

        self._reporter.report_result(result)
        if not result.success:
            raise _Exit(1)

    def run(self) -> int:
        try:
            self._run()
        except _Exit as e:
            return e.code
        return 0



from __future__ import annotations

import io
import json
import sys
import typing as t

import ruamel.yaml

from ..cachedownloader import CacheDownloader
from ..parsers import ParseError, ParserSet
from ..utils import filename2path
from .errors import SchemaParseError

yaml = ruamel.yaml.YAML(typ="safe")


class _UnsetType:
    pass


_UNSET = _UnsetType()


def _run_load_callback(schema_location: str, callback: t.Callable) -> dict:
    try:
        schema = callback()
    # only local loads can raise the YAMLError, but catch for both cases for simplicity
    except (ValueError, ruamel.yaml.error.YAMLError) as e:
        raise SchemaParseError(schema_location) from e
    if not isinstance(schema, dict):
        raise SchemaParseError(schema_location)
    return schema


class LocalSchemaReader:
    def __init__(self, filename: str) -> None:
        self.path = filename2path(filename)
        self.filename = str(self.path)
        self.parsers = ParserSet()
        self._parsed_schema: dict | _UnsetType = _UNSET

    def get_retrieval_uri(self) -> str | None:
        return self.path.as_uri()

    def _read_impl(self) -> t.Any:
        return self.parsers.parse_file(self.path, default_filetype="json")

    def read_schema(self) -> dict:
        if self._parsed_schema is _UNSET:
            self._parsed_schema = _run_load_callback(self.filename, self._read_impl)
        return t.cast(dict, self._parsed_schema)


class StdinSchemaReader:
    def __init__(self) -> None:
        self.parsers = ParserSet()
        self._parsed_schema: dict | _UnsetType = _UNSET

    def get_retrieval_uri(self) -> str | None:
        return None

    def read_schema(self) -> dict:
        if self._parsed_schema is _UNSET:
            try:
                self._parsed_schema = json.load(sys.stdin)
            except ValueError as e:
                raise ParseError("Failed to parse JSON from stdin") from e
        return t.cast(dict, self._parsed_schema)


class HttpSchemaReader:
    def __init__(
        self,
        url: str,
        cache_filename: str | None,
        disable_cache: bool,
    ) -> None:
        self.url = url
        self.parsers = ParserSet()
        self.downloader = CacheDownloader(
            url,
            cache_filename,
            disable_cache=disable_cache,
            validation_callback=self._parse,
        )
        self._parsed_schema: dict | _UnsetType = _UNSET

    def _parse(self, schema_bytes: bytes) -> t.Any:
        return self.parsers.parse_data_with_path(
            io.BytesIO(schema_bytes), self.url, default_filetype="json"
        )

    def get_retrieval_uri(self) -> str | None:
        return self.url

    def _read_impl(self) -> t.Any:
        with self.downloader.open() as fp:
            return self._parse(fp.read())

    def read_schema(self) -> dict:
        if self._parsed_schema is _UNSET:
            self._parsed_schema = _run_load_callback(self.url, self._read_impl)
        return t.cast(dict, self._parsed_schema)



import json
import os
import platform
import time

import pytest
import requests
import responses

from check_jsonschema.cachedownloader import CacheDownloader, FailedDownloadError


def add_default_response():
    responses.add(
        "GET",
        "https://example.com/schema1.json",
        headers={"Last-Modified": "Sun, 01 Jan 2000 00:00:01 GMT"},
        json={},
        match_querystring=None,
    )


@pytest.fixture
def default_response():
    add_default_response()


def test_default_filename_from_uri(default_response):
    cd = CacheDownloader("https://example.com/schema1.json")
    assert cd._filename == "schema1.json"


@pytest.mark.parametrize(
    "sysname, fakeenv, expect_value",
    [
        ("Windows", {}, None),
        (
            "Windows",
            {"LOCALAPPDATA": "localappdata", "APPDATA": "appdata"},
            "localappdata",
        ),
        ("Windows", {"LOCALAPPDATA": "localappdata"}, "localappdata"),
        ("Windows", {"APPDATA": "appdata"}, "appdata"),
        ("Darwin", {}, "<expanduser>"),
        ("Linux", {}, "<expanduser>"),
        ("Linux", {"XDG_CACHE_HOME": "xdg-cache"}, "xdg-cache"),
    ],
)
def test_default_cache_dir(
    monkeypatch, default_response, sysname, fakeenv, expect_value
):
    for var in ["LOCALAPPDATA", "APPDATA", "XDG_CACHE_HOME"]:
        monkeypatch.delenv(var, raising=False)
    for k, v in fakeenv.items():
        monkeypatch.setenv(k, v)
    if expect_value is not None:
        expect_value = os.path.join(expect_value, "check_jsonschema", "downloads")

    def fakesystem():
        return sysname

    expanduser_path = None

    def fake_expanduser(path):
        nonlocal expanduser_path
        expanduser_path = path
        return "<expanduser>"

    monkeypatch.setattr(platform, "system", fakesystem)
    monkeypatch.setattr(os.path, "expanduser", fake_expanduser)

    cd = CacheDownloader("https://example.com/schema1.json")
    assert cd._cache_dir == expect_value

    if sysname == "Darwin":
        assert expanduser_path == "~/Library/Caches"
    elif sysname == "Linux":
        assert expanduser_path == "~/.cache"
    else:
        assert expanduser_path is None


def test_cache_hit_by_mtime(monkeypatch, default_response):
    monkeypatch.setattr(os.path, "exists", lambda x: True)

    # local mtime = NOW, cache hit
    monkeypatch.setattr(os.path, "getmtime", lambda x: time.time())
    cd = CacheDownloader("https://example.com/schema1.json")
    assert cd._cache_hit(
        "/tmp/schema1.json",
        requests.get("https://example.com/schema1.json", stream=True),
    )

    # local mtime = 0, cache miss
    monkeypatch.setattr(os.path, "getmtime", lambda x: 0)
    cd = CacheDownloader("https://example.com/schema1.json")
    assert (
        cd._cache_hit(
            "/tmp/schema1.json",
            requests.get("https://example.com/schema1.json", stream=True),
        )
        is False
    )


def test_cachedownloader_cached_file(tmp_path, monkeypatch, default_response):
    # create a file
    f = tmp_path / "foo.json"
    f.write_text("{}")

    # set the cache_dir to the tmp dir (so that cache_dir will always be set)
    cd = CacheDownloader(str(f), cache_dir=tmp_path)
    # patch the downloader to skip any download "work"
    monkeypatch.setattr(cd, "_download", lambda: str(f))

    with cd.open() as fp:
        assert fp.read() == b"{}"


@pytest.mark.parametrize(
    "mode", ["filename", "filename_otherdir", "cache_dir", "disable_cache"]
)
@pytest.mark.parametrize("failures", (0, 1, 10, requests.ConnectionError))
def test_cachedownloader_e2e(tmp_path, mode, failures):
    if isinstance(failures, int):
        for _i in range(failures):
            responses.add(
                "GET",
                "https://example.com/schema1.json",
                status=500,
                match_querystring=None,
            )
    else:
        responses.add(
            "GET",
            "https://example.com/schema1.json",
            body=failures(),
            match_querystring=None,
        )
    add_default_response()
    f = tmp_path / "schema1.json"
    if mode == "filename":
        cd = CacheDownloader(
            "https://example.com/schema1.json", filename=str(f), cache_dir=str(tmp_path)
        )
    elif mode == "filename_otherdir":
        otherdir = tmp_path / "otherdir"
        cd = CacheDownloader(
            "https://example.com/schema1.json", filename=str(f), cache_dir=str(otherdir)
        )
    elif mode == "cache_dir":
        cd = CacheDownloader(
            "https://example.com/schema1.json", cache_dir=str(tmp_path)
        )
    elif mode == "disable_cache":
        cd = CacheDownloader("https://example.com/schema1.json", disable_cache=True)
    else:
        raise NotImplementedError

    if isinstance(failures, int) and failures < 3:
        with cd.open() as fp:
            assert fp.read() == b"{}"
        if mode == "filename":
            assert f.exists()
        elif mode == "filename_otherdir":
            otherdir = f.exists()
        elif mode == "cache_dir":
            assert (tmp_path / "schema1.json").exists()
        elif mode == "disable_cache":
            assert not (tmp_path / "schema1.json").exists()
            assert not f.exists()
        else:
            raise NotImplementedError
    else:
        with pytest.raises(FailedDownloadError):
            with cd.open() as fp:
                pass
        assert not (tmp_path / "schema1.json").exists()
        assert not f.exists()


@pytest.mark.parametrize("disable_cache", (True, False))
def test_cachedownloader_retries_on_bad_data(tmp_path, disable_cache):
    responses.add(
        "GET",
        "https://example.com/schema1.json",
        status=200,
        body="{",
        match_querystring=None,
    )
    add_default_response()
    f = tmp_path / "schema1.json"
    cd = CacheDownloader(
        "https://example.com/schema1.json",
        filename=str(f),
        cache_dir=str(tmp_path),
        disable_cache=disable_cache,
        validation_callback=json.loads,
    )

    with cd.open() as fp:
        assert fp.read() == b"{}"

    if disable_cache:
        assert not f.exists()
    else:
        assert f.exists()


@pytest.mark.parametrize("file_exists", (True, False))
@pytest.mark.parametrize(
    "failure_mode", ("header_missing", "header_malformed", "time_overflow")
)
def test_cachedownloader_handles_bad_lastmod_header(
    monkeypatch, tmp_path, file_exists, failure_mode
):
    if failure_mode == "header_missing":
        responses.add(
            "GET",
            "https://example.com/schema1.json",
            headers={},
            json={},
            match_querystring=None,
        )
    elif failure_mode == "header_malformed":
        responses.add(
            "GET",
            "https://example.com/schema1.json",
            headers={"Last-Modified": "Jan 2000 00:00:01"},
            json={},
            match_querystring=None,
        )
    elif failure_mode == "time_overflow":
        add_default_response()

        def fake_mktime(*args):
            raise OverflowError("uh-oh")

        monkeypatch.setattr("time.mktime", fake_mktime)
    else:
        raise NotImplementedError

    original_file_contents = b'{"foo": "bar"}'
    f = tmp_path / "schema1.json"

    if file_exists:
        f.write_bytes(original_file_contents)
    else:
        assert not f.exists()

    cd = CacheDownloader(
        "https://example.com/schema1.json", filename=str(f), cache_dir=str(tmp_path)
    )

    # if the file already existed, it will not be overwritten by the cachedownloader
    # so the returned value for both the downloader and a direct file read should be the
    # original contents
    if file_exists:
        with cd.open() as fp:
            assert fp.read() == original_file_contents
        assert f.read_bytes() == original_file_contents
    # otherwise, the file will have been created with new content
    # both reads will show that new content
    else:
        with cd.open() as fp:
            assert fp.read() == b"{}"
        assert f.read_bytes() == b"{}"

    # at the end, the file always exists on disk
    assert f.exists()
